"""
To run tests:
    python -m tests.test_connect
"""
import array as ar
import configparser
from datetime import datetime
import os
import sys
import unittest

try:
    from unittest.mock import patch
except:
    from mock import patch  # py2

from _connect import ConnectError as _ConnectError
from chirpsdk import (
    CHIRP_CONNECT_STATE_STOPPED,
    CHIRP_CONNECT_STATE_PAUSED,
    CHIRP_CONNECT_STATE_RUNNING,
    CHIRP_CONNECT_STATE_SENDING,
    CHIRP_CONNECT_STATE_NOT_CREATED,
    AudioSet, ChirpConnect, ConnectError
)


class TestConnectSDK(unittest.TestCase):
    BUFFER_SIZE = 1024
    TEST_PAYLOAD_LENGTH = 3

    def setUp(self):
        config = configparser.ConfigParser()
        config.read(os.path.expanduser('~/.chirprc'))

        self.async_request = patch('requests_futures.sessions.FuturesSession.post')
        self.async_patch = self.async_request.start()
        self.addCleanup(self.async_request.stop)

        try:
            self.app_key = config.get('test', 'app_key')
            self.app_secret = config.get('test', 'app_secret')
            self.app_config = config.get('test', 'app_config')
        except configparser.NoSectionError:
            raise Exception("Couldn't find test credentials. Please add a [test] section to your ~/.chirprc.")

        self.sdk = ChirpConnect(self.app_key, self.app_secret, self.app_config)
        self.sdk.audio = AudioSet(self.sdk)
        self.is2 = sys.version[0] == '2'

        self.length = None
        self.channel = None

    def tearDown(self):
        self.sdk.audio.stop()
        self.sdk.audio.close()
        del self.sdk

    def test_version(self):
        version = self.sdk.version
        for module in ['connect', 'core', 'crypto']:
            self.assertIn(module, version)
            for data in ['name', 'version', 'build']:
                self.assertIn(data, version[module])
                self.assertTrue(len(version[module][data]) > 0)

    # -- Getters & Setters

    def test_volume(self):
        self.assertEqual(self.sdk.volume, 1.0)

    def test_set_volume(self):
        self.sdk.volume = 0.33
        self.assertEqual(self.sdk.volume, 0.33)

    def test_sample_rate(self):
        self.assertEqual(self.sdk.sample_rate, 44100)

    def test_set_sample_rate(self):
        self.sdk.sample_rate = 48000
        self.assertEqual(self.sdk.sample_rate, 48000)
        with self.assertRaises(ConnectError):
            self.sdk.sample_rate = 0

    def test_default_state(self):
        self.assertEqual(self.sdk.state, CHIRP_CONNECT_STATE_STOPPED)

    def test_get_auto_mute(self):
        self.assertTrue(self.sdk.auto_mute)

    def test_set_auto_mute(self):
        self.sdk.auto_mute = False
        self.assertFalse(self.sdk.auto_mute)

    def test_protocol_name(self):
        self.assertEqual(self.sdk.protocol_name.decode(), 'standard')

    def test_protocol_version(self):
        self.assertIsInstance(self.sdk.protocol_version, int)

    def test_protocol_duration(self):
        self.assertEqual(self.sdk.get_duration(10), 2.04)

    def test_expiry(self):
        self.assertIsInstance(self.sdk.expiry, datetime)

    def test_channel_count(self):
        self.assertEqual(self.sdk.channel_count, 1)

    def test_transmission_channel(self):
        self.assertEqual(self.sdk.transmission_channel, 0)

    def test_get_state_for_channel(self):
        self.sdk.start()
        self.assertEqual(self.sdk.get_state_for_channel(0), CHIRP_CONNECT_STATE_RUNNING)
        payload = self.sdk.random_payload(self.TEST_PAYLOAD_LENGTH)
        self.sdk.send(payload)
        self.assertEqual(self.sdk.get_state_for_channel(0), CHIRP_CONNECT_STATE_SENDING)

    # -- States

    def test_not_created(self):
        with self.assertRaises(_ConnectError):
            sdk = ChirpConnect(self.app_key, self.app_secret, '/not/real/path')
            self.assertEqual(sdk.state, CHIRP_CONNECT_STATE_NOT_CREATED)

    def test_start(self):
        self.sdk.start()
        self.assertEqual(self.sdk.state, CHIRP_CONNECT_STATE_RUNNING)

    def test_already_started(self):
        self.sdk.start()
        with self.assertRaises(ConnectError):
            self.sdk.start()

    def test_pause(self):
        self.sdk.start()
        self.sdk.pause(True)
        self.assertEqual(self.sdk.state, CHIRP_CONNECT_STATE_PAUSED)

    def test_unpause(self):
        self.sdk.start()
        self.sdk.pause(True)
        self.sdk.pause(False)
        self.assertEqual(self.sdk.state, CHIRP_CONNECT_STATE_RUNNING)

    def test_already_paused(self):
        with self.assertRaises(ConnectError):
            self.sdk.pause(True)

    def test_stop(self):
        self.sdk.start()
        self.sdk.stop()
        self.assertEqual(self.sdk.state, CHIRP_CONNECT_STATE_STOPPED)

    def test_already_stopped(self):
        with self.assertRaises(ConnectError):
            self.sdk.stop()

    # -- Callbacks

    def stub_connect_state_callback(self, old, new):
        self.old = old
        self.new = new

    def stub_connect_callback(self, payload, channel):
        self.length = len(payload)
        self.channel = channel

    def stub_receiving_callback(self, channel):
        self.recv = True
        self.channel = channel

    def test_state_changed_callback(self):
        self.sdk.callbacks.on_state_changed = self.stub_connect_state_callback
        self.sdk.trigger_callbacks([0, 1, 2, 3, 4])
        self.assertIsNotNone(self.old)
        self.assertIsNotNone(self.new)

    def test_sending_callback(self):
        self.sdk.callbacks.on_sending = self.stub_connect_callback
        self.sdk.trigger_callbacks([0, 1, 2, 3, 4])
        self.assertIsNotNone(self.length)
        self.assertIsNotNone(self.channel)

    def test_sent_callback(self):
        self.sdk.callbacks.on_sent = self.stub_connect_callback
        self.sdk.trigger_callbacks([0, 1, 2, 3, 4])
        self.assertIsNotNone(self.length)
        self.assertIsNotNone(self.channel)

    def test_receiving_callback(self):
        self.sdk.callbacks.on_receiving = self.stub_receiving_callback
        self.sdk.trigger_callbacks([0, 1, 2, 3, 4])
        self.assertTrue(self.recv)
        self.assertIsNotNone(self.channel)

    def test_received_callback(self):
        self.sdk.callbacks.on_received = self.stub_connect_callback
        self.sdk.trigger_callbacks([0, 1, 2, 3, 4])
        self.assertIsNotNone(self.length)
        self.assertIsNotNone(self.channel)

    # -- Processing

    def test_process_input(self):
        indata = ar.array('f', [0.025] * self.BUFFER_SIZE)
        self.sdk.start()
        self.sdk.process_input(getattr(indata, 'tostring' if self.is2 else 'tobytes')())

    def test_process_input_not_started(self):
        indata = ar.array('f', [0.025] * self.BUFFER_SIZE)
        with self.assertRaises(_ConnectError):
            self.sdk.process_input(getattr(indata, 'tostring' if self.is2 else 'tobytes')())

    def test_process_output(self):
        outdata = ar.array('f', [0.05] * self.BUFFER_SIZE)
        self.sdk.start()
        self.sdk.process_output(getattr(outdata, 'tostring' if self.is2 else 'tobytes')())

    def test_process_output_not_started(self):
        outdata = ar.array('f', [0.05] * self.BUFFER_SIZE)
        with self.assertRaises(_ConnectError):
            self.sdk.process_output(getattr(outdata, 'tostring' if self.is2 else 'tobytes')())

    def test_process_shorts_input(self):
        indata = ar.array('h', [128] * self.BUFFER_SIZE)
        self.sdk.start()
        self.sdk.process_shorts_input(getattr(indata, 'tostring' if self.is2 else 'tobytes')())

    def test_process_shorts_output(self):
        outdata = ar.array('h', [-128] * self.BUFFER_SIZE)
        self.sdk.start()
        self.sdk.process_shorts_output(getattr(outdata, 'tostring' if self.is2 else 'tobytes')())

    # -- Payload

    def test_get_max_payload_length(self):
        self.assertIsInstance(self.sdk.max_payload_length, int)
        self.assertTrue(self.sdk.max_payload_length > 0)

    def test_new_payload_string(self):
        payload = self.sdk.new_payload('test'.encode())
        self.assertIsInstance(payload, bytearray)

    def test_new_payload_array(self):
        payload = self.sdk.new_payload([64, 27, 33, 27])
        self.assertIsInstance(payload, bytearray)

    def test_random_payload(self):
        payload = self.sdk.random_payload(self.TEST_PAYLOAD_LENGTH)
        self.assertIsInstance(payload, bytearray)
        for byte in range(0, len(payload)):
            self.assertIsInstance(payload[byte], int)

    def test_pseudo_random_payload(self):
        self.sdk._set_seed(0)
        self.assertEqual(
            self.sdk.random_payload(0),
            [47, 117, 192, 67, 251, 195, 103, 9, 211,
             21, 242, 36, 87, 70, 216, 88, 140, 58, 193])

    def test_is_valid(self):
        payload = self.sdk.random_payload(self.TEST_PAYLOAD_LENGTH)
        self.assertTrue(self.sdk.is_valid(payload))

    def test_payload_is_valid(self):
        payload = self.sdk.random_payload(self.TEST_PAYLOAD_LENGTH)
        self.assertTrue(payload.isvalid())

    def test_as_string(self):
        payload = self.sdk.random_payload(self.TEST_PAYLOAD_LENGTH)
        self.assertIsInstance(self.sdk.as_string(payload), str)

    def test_payload_as_string(self):
        payload = self.sdk.random_payload(self.TEST_PAYLOAD_LENGTH)
        self.assertIsInstance(str(payload), str)

    def test_send(self):
        self.sdk.start()
        payload = self.sdk.random_payload(self.TEST_PAYLOAD_LENGTH)
        self.assertIsNone(self.sdk.send(payload))

    def test_null_payload(self):
        with self.assertRaises(ValueError):
            self.sdk.new_payload([])

    def test_payload_too_long(self):
        payload = self.sdk.new_payload('hello'.encode('ascii'))
        with self.assertRaises(ValueError):
            payload.extend('this-is-wayyyyy-toooooo-long')


if __name__ == '__main__':
    unittest.main()



# ------------------------------------------------------------------------
#
#  This file is part of the Chirp Connect Python SDK.
#  For full information on usage and licensing, see https://chirp.io/
#
#  Copyright (c) 2011-2018, Asio Ltd.
#  All rights reserved.
#
# ------------------------------------------------------------------------
from datetime import datetime

import requests
from requests_futures.sessions import FuturesSession

from . import __version__
from .exceptions import ConnectNetworkError

session = FuturesSession()
config_url = 'https://config.chirp.io'
analytics_url = 'https://analytics.chirp.io'


def get_config_from_network(key, secret, name=None):
    """
    Retrieve a config for an application.
    If name is specified then this config is returned, otherwise
    the default is used.

    Args:
        name (str): config name, default is used if not set

    Returns: (str) config string

    Raises:
        ConnectNetworkError: If config request fails
        ValueError: If an invalid config name is requested
    """
    try:
        url = config_url + '/v3/connect'
        r = requests.get(url, auth=(key, secret))
        if r.status_code != 200:
            err = r.json()
            raise ConnectNetworkError(err.get('message', 'Failed to retrieve config'))
        data = r.json()
        configs = [l['name'] for l in data['data']]
        if name and name not in configs:
            raise ValueError('Invalid config name')
        return data['data'][configs.index(name)]['config'] if name else data['data'][0]['config']
    except requests.exceptions.ConnectionError:
        raise ConnectNetworkError('No internet connection')
    except requests.exceptions.Timeout:
        raise ConnectNetworkError('Timeout')
    except requests.exceptions.RequestException as err:
        raise ConnectNetworkError(str(err))


def async_request(url, auth, data):
    """ Asynschronous request with no consequences """
    try:
        session.post(url, auth=auth, json=data)
    except requests.exceptions.ConnectionError:
        pass
    except requests.exceptions.Timeout:
        pass
    except requests.exceptions.RequestException as err:
        pass


def create_instantiate(key, secret, uid):
    """ Instantiated the SDK """
    url = analytics_url + '/v3/connect/instantiate'
    async_request(url, (key, secret), {
        'client_id': uid,
        'timestamp': datetime.now().isoformat(),
        'platform': 'python',
        'sdk_version': __version__
    })


def create_send(key, secret, uid, length, protocol_name, protocol_version):
    """ Sent a payload """
    url = analytics_url + '/v3/connect/send'
    async_request(url, (key, secret), {
        'client_id': uid,
        'timestamp': datetime.now().isoformat(),
        'payload_length': length,
        'protocol_name': protocol_name.decode(),
        'protocol_version': protocol_version,
        'platform': 'python',
        'sdk_version': __version__
    })


def create_receive(key, secret, uid, length, protocol_name, protocol_version):
    """ Received a payload """
    url = analytics_url + '/v3/connect/receive'
    async_request(url, (key, secret), {
        'client_id': uid,
        'timestamp': datetime.now().isoformat(),
        'success': length != 0,
        'payload_length': length,
        'protocol_name': protocol_name.decode(),
        'protocol_version': protocol_version,
        'platform': 'python',
        'sdk_version': __version__
    })



# ------------------------------------------------------------------------
#
#  This file is part of the Chirp Connect Python SDK.
#  For full information on usage and licensing, see https://chirp.io/
#
#  Copyright (c) 2011-2018, Asio Ltd.
#  All rights reserved.
#
# ------------------------------------------------------------------------

import platform
import time
import threading

try:
    from multiprocessing import SimpleQueue
except ImportError:  # python2
    from multiprocessing.queues import SimpleQueue

import sounddevice as sd

from .exceptions import ConnectAudioError


class Audio(object):
    """
    Chirp Connect Audio I/O
    """
    def __init__(self, sdk):
        self.sdk = sdk
        self.processing = None
        self.input_stream = None
        self.output_stream = None

        self.output_channels = 1
        self.wav_filename = None
        self.sample_size = 'float32'
        self.block_size = 4096 if platform.system() == 'Linux' else 0

        if self.sample_size == 'float32':
            self.sample_format = 'f'
            self.process_input_fn = self.sdk.process_input
            self.process_output_fn = self.sdk.process_output
        elif self.sample_size == 'int16':
            self.sample_format = 'h'
            self.process_input_fn = self.sdk.process_shorts_input
            self.process_output_fn = self.sdk.process_shorts_output
        else:
            raise ConnectAudioError(
                'Invalid audio format: %s. Only float32 and int16 are supported' %
                self.sample_size)

    def start(self, send=True, receive=True):
        """ Start audio I/O stream and processing thread """
        if receive:
            sd.check_input_settings(
                device=self.input_device, channels=1,
                dtype=self.sample_size, samplerate=self.sdk.sample_rate
            )
            self.processing = AudioProcessingThread(parent=self)
            self.input_stream = sd.RawInputStream(
                device=self.input_device,
                channels=1,
                samplerate=int(self.sdk.sample_rate),
                dtype=self.sample_size,
                blocksize=self.block_size,
                callback=self.process_input
            )
            self.input_stream.start()
        if send:
            sd.check_output_settings(
                device=self.output_device, channels=self.output_channels,
                dtype=self.sample_size, samplerate=self.sdk.sample_rate
            )
            self.output_stream = sd.RawOutputStream(
                device=self.output_device,
                channels=self.output_channels,
                samplerate=int(self.sdk.sample_rate),
                dtype=self.sample_size,
                blocksize=self.block_size,
                callback=self.process_output
            )
            self.output_stream.start()

    def stop(self):
        if self.processing:
            self.processing.stop()
        if self.input_stream:
            self.input_stream.stop()
        if self.output_stream:
            self.output_stream.stop()

    def process_input(self, indata, frames, time, status):
        """ Process input audio data - receive only audio callback """
        self.processing.block_size = frames
        self.processing.input_queue.put(bytes(indata))

    def process_output(self, outdata, frames, time, status):
        """ Process output audio data - send only audio callback """
        self.process_output_fn(outdata)

    def query_devices(self):
        """ Return info about available devices """
        return sd.query_devices()

    @property
    def input_device(self):
        return sd.default.device[0]

    @input_device.setter
    def input_device(self, index):
        sd.default.device[0] = index

    @property
    def output_device(self):
        return sd.default.device[1]

    @output_device.setter
    def output_device(self, index):
        sd.default.device[1] = index

    def close(self):
        """ Close I/O streams """
        if self.input_stream:
            self.input_stream.close()
        if self.output_stream:
            self.output_stream.close()


class AudioProcessingThread(threading.Thread):
    """
    Chirp Connect audio processing thread
    """
    DEBUG_AUDIO_FILENAME = 'chirp_audio.wav'

    def __init__(self, parent=None, *args, **kwargs):
        """
        Initialise audio processing.
        In debug mode, the audio data is saved to file.
        """
        self.sdk = parent.sdk
        self.sample_size = parent.sample_size
        self.block_size = parent.block_size
        self.sample_format = parent.sample_format
        self.process_input_fn = parent.process_input_fn
        self.sample_rate = float(parent.sdk.sample_rate)

        self.block_period = self.block_size / self.sample_rate or 0.1
        self.wav_filename = parent.wav_filename or self.DEBUG_AUDIO_FILENAME
        self.input_queue = SimpleQueue()
        super(AudioProcessingThread, self).__init__(*args, **kwargs)

        if self.sdk.debug:
            import soundfile as sf
            self.wav_file = sf.SoundFile(
                self.wav_filename, mode='w', channels=1,
                samplerate=self.sdk.sample_rate)

        self.daemon = True
        self.start()

    def run(self):
        """
        Continuously process any input data from circular buffer.
        Note: We need to sleep as much as possible in this thread
        to restrict CPU usage.
        """
        while self.is_alive():

            tstart = time.time()
            while not self.input_queue.empty():
                data = self.input_queue.get()
                self.process_input_fn(data)
                if self.sdk.debug and not self.wav_file.closed:
                    self.wav_file.buffer_write(data, dtype=self.sample_size)
                self.block_period = self.block_size / self.sample_rate

            tsleep = (self.block_period - ((time.time() - tstart)))
            if tsleep > 0:
                time.sleep(tsleep)

    def stop(self):
        """ In debug mode, close wav file """
        if self.sdk.debug:
            self.wav_file.close()



# ------------------------------------------------------------------------
#
#  This file is part of the Chirp Connect Python SDK.
#  For full information on usage and licensing, see https://chirp.io/
#
#  Copyright (c) 2011-2018, Asio Ltd.
#  All rights reserved.
#
# ------------------------------------------------------------------------


class ConnectError(Exception):

    def __init__(self, message, code=None):
        Exception.__init__(self)
        self.message = message
        self.error_code = code

    def __str__(self):
        return self.message


class ConnectAudioError(Exception):

    pass


class ConnectNetworkError(Exception):

    pass



# ------------------------------------------------------------------------
#
#  This file is part of the Chirp Connect Python SDK.
#  For full information on usage and licensing, see https://chirp.io/
#
#  Copyright (c) 2011-2018, Asio Ltd.
#  All rights reserved.
#
# ------------------------------------------------------------------------
import os
import uuid
from . import libconnect, __version__

import configparser
from ctypes import (
    Structure, CFUNCTYPE, POINTER, pointer,
    c_char_p, c_uint8, c_uint32, c_uint64,
    c_short, c_float, c_void_p, c_bool, c_size_t
)
from datetime import datetime
from platform import architecture
from random import randrange

from .audio import Audio
from .exceptions import ConnectError
from .network import (
    get_config_from_network,
    create_instantiate,
    create_send,
    create_receive
)
from _connect import (
    new_chirp_connect,
    del_chirp_connect,
    chirp_connect_set_config,
    chirp_connect_process_input,
    chirp_connect_process_output,
    chirp_connect_process_shorts_input,
    chirp_connect_process_shorts_output,
    CHIRP_CONNECT_STATE_STOPPED,
    CHIRP_CONNECT_STATE_PAUSED,
    CHIRP_CONNECT_STATE_RUNNING,
    CHIRP_CONNECT_STATE_SENDING,
    CHIRP_CONNECT_STATE_RECEIVING,
    CHIRP_CONNECT_STATE_NOT_CREATED,
)
from _connect import ConnectError as _ConnectError

CHIRP_CONNECT_STATE = {
    CHIRP_CONNECT_STATE_STOPPED: 'Stopped',
    CHIRP_CONNECT_STATE_PAUSED: 'Paused',
    CHIRP_CONNECT_STATE_RUNNING: 'Running',
    CHIRP_CONNECT_STATE_SENDING: 'Sending',
    CHIRP_CONNECT_STATE_RECEIVING: 'Receiving',
    CHIRP_CONNECT_STATE_NOT_CREATED: 'Not Created'
}

# ------------------------------------------------------------------------
# Callback function pointers
# ------------------------------------------------------------------------
ChirpConnectCallback = CFUNCTYPE(None, c_void_p, POINTER(c_uint8), c_size_t, c_uint8)
ChirpConnectStateCallback = CFUNCTYPE(None, c_void_p, c_uint8, c_uint8)


class ChirpPackage(Structure):
    _fields_ = [
        ("name", c_char_p),
        ("version", c_char_p),
        ("build", c_char_p)
    ]


class ChirpVersion(Structure):
    _fields_ = [
        ("connect", ChirpPackage),
        ("core", ChirpPackage),
        ("crypto", ChirpPackage)
    ]


class ChirpConnectCallbackSet(Structure):
    _fields_ = [
        ("on_state_changed", ChirpConnectStateCallback),
        ("on_sending", ChirpConnectCallback),
        ("on_sent", ChirpConnectCallback),
        ("on_receiving", ChirpConnectCallback),
        ("on_received", ChirpConnectCallback)
    ]

    def __init__(self, sdk):
        self.sdk = sdk
        self.state_changed_fn = ChirpConnectStateCallback(self.state_changed)
        self.sending_fn = ChirpConnectCallback(self.sending)
        self.sent_fn = ChirpConnectCallback(self.sent)
        self.receiving_fn = ChirpConnectCallback(self.receiving)
        self.received_fn = ChirpConnectCallback(self.received)
        super(ChirpConnectCallbackSet, self).__init__(
            self.state_changed_fn, self.sending_fn, self.sent_fn,
            self.receiving_fn, self.received_fn
        )

    def state_changed(self, cptr, old, new):
        self.sdk.callbacks.on_state_changed(old, new)

    def sending(self, cptr, payload, length, channel):
        self.sdk.callbacks.on_sending(
            _Payload(self.sdk, payload[:length]) if length else None, channel)

    def sent(self, cptr, payload, length, channel):
        self.sdk.callbacks.on_sent(
            _Payload(self.sdk, payload[:length]) if length else None, channel)

    def receiving(self, cptr, payload, length, channel):
        self.sdk.callbacks.on_receiving(channel)

    def received(self, cptr, payload, length, channel):
        self.sdk.callbacks.on_received(
            _Payload(self.sdk, payload[:length]) if length else None, channel)
        if not self.sdk.is_offline:
            create_receive(
                self.sdk.key, self.sdk.secret, self.sdk._uid,
                length, self.sdk.protocol_name, self.sdk.protocol_version)


# ------------------------------------------------------------------------
# Symbol bindings: Utils
# ------------------------------------------------------------------------
libconnect.chirp_connect_get_chirp_versions.restype = ChirpVersion
libconnect.chirp_connect_get_package_full_version.restype = c_char_p
libconnect.chirp_connect_get_package_name.restype = c_char_p
libconnect.chirp_connect_get_package_version.restype = c_char_p
libconnect.chirp_connect_get_package_build.restype = c_char_p
libconnect.chirp_connect_error_code_to_string.argtypes = [c_uint8]
libconnect.chirp_connect_error_code_to_string.restype = c_char_p
libconnect.chirp_connect_trigger_callbacks.argtypes = [c_void_p, POINTER(c_uint8), c_size_t]
libconnect.chirp_connect_is_offline_mode.argtypes = [c_void_p]
libconnect.chirp_connect_is_offline_mode.restype = c_bool

# ------------------------------------------------------------------------
# Symbol bindings: Getters & Setters
# ------------------------------------------------------------------------
libconnect.chirp_connect_set_config.argtypes = [c_void_p, c_char_p]
libconnect.chirp_connect_set_config.restype = c_uint8
libconnect.chirp_connect_set_random_seed.argtypes = [c_void_p, c_uint32]
libconnect.chirp_connect_set_random_seed.restype = c_uint8
libconnect.chirp_connect_set_volume.argtypes = [c_void_p, c_float]
libconnect.chirp_connect_set_volume.restype = c_uint8
libconnect.chirp_connect_get_volume.argtypes = [c_void_p]
libconnect.chirp_connect_get_volume.restype = c_float
libconnect.chirp_connect_set_sample_rate.argtypes = [c_void_p, c_uint32]
libconnect.chirp_connect_set_sample_rate.restype = c_uint8
libconnect.chirp_connect_get_sample_rate.argtypes = [c_void_p]
libconnect.chirp_connect_get_sample_rate.restype = c_uint32
libconnect.chirp_connect_get_state.argtypes = [c_void_p]
libconnect.chirp_connect_get_state.restype = c_uint8
libconnect.chirp_connect_set_auto_mute.argtypes = [c_void_p, c_bool]
libconnect.chirp_connect_set_auto_mute.restype = c_uint8
libconnect.chirp_connect_get_auto_mute.argtypes = [c_void_p]
libconnect.chirp_connect_get_auto_mute.restype = c_bool
libconnect.chirp_connect_get_protocol_name.argtypes = [c_void_p]
libconnect.chirp_connect_get_protocol_name.restype = c_char_p
libconnect.chirp_connect_get_duration_for_payload_length.argtypes = [c_void_p, c_size_t]
libconnect.chirp_connect_get_duration_for_payload_length.restype = c_float
libconnect.chirp_connect_get_protocol_version.argtypes = [c_void_p]
libconnect.chirp_connect_get_protocol_version.restype = c_uint8
libconnect.chirp_connect_get_expiry_time.argtypes = [c_void_p]
if architecture()[0] == '32bit':  # Determine size of time_t
    libconnect.chirp_connect_get_expiry_time.restype = c_uint32
else:
    libconnect.chirp_connect_get_expiry_time.restype = c_uint64
libconnect.chirp_connect_get_state_for_channel.argtypes = [c_void_p, c_uint8]
libconnect.chirp_connect_get_state_for_channel.restype = c_uint8
libconnect.chirp_connect_get_transmission_channel.argtypes = [c_void_p]
libconnect.chirp_connect_get_transmission_channel.restype = c_uint8
libconnect.chirp_connect_set_transmission_channel.argtypes = [c_void_p, c_uint8]
libconnect.chirp_connect_set_transmission_channel.restype = c_uint8
libconnect.chirp_connect_get_channel_count.argtypes = [c_void_p]
libconnect.chirp_connect_get_channel_count.restype = c_uint8

# ------------------------------------------------------------------------
# Symbol bindings: States
# ------------------------------------------------------------------------
libconnect.chirp_connect_start.argtypes = [c_void_p]
libconnect.chirp_connect_start.restype = c_uint8
libconnect.chirp_connect_pause.argtypes = [c_void_p, c_bool]
libconnect.chirp_connect_pause.restype = c_uint8
libconnect.chirp_connect_stop.argtypes = [c_void_p]
libconnect.chirp_connect_stop.restype = c_uint8

# ------------------------------------------------------------------------
# Symbol bindings: Callbacks
# ------------------------------------------------------------------------
libconnect.chirp_connect_set_callbacks.argtypes = [c_void_p, ChirpConnectCallbackSet]
libconnect.chirp_connect_set_callbacks.restype = c_uint8
libconnect.chirp_connect_set_callback_ptr.argtypes = [c_void_p, c_void_p]
libconnect.chirp_connect_set_callback_ptr.restype = c_uint8

# ------------------------------------------------------------------------
# Symbol bindings: Payload
# ------------------------------------------------------------------------
libconnect.chirp_connect_get_max_payload_length.argtypes = [c_void_p]
libconnect.chirp_connect_get_max_payload_length.restype = c_size_t
libconnect.chirp_connect_new_payload.argtypes = [c_void_p, c_size_t]
libconnect.chirp_connect_new_payload.restype = POINTER(c_uint8)
libconnect.chirp_connect_random_payload.argtypes = [c_void_p, POINTER(c_size_t)]
libconnect.chirp_connect_random_payload.restype = POINTER(c_uint8)
libconnect.chirp_connect_is_valid.argtypes = [c_void_p, POINTER(c_uint8), c_size_t]
libconnect.chirp_connect_is_valid.restype = c_uint8
libconnect.chirp_connect_as_string.argtypes = [c_void_p, POINTER(c_uint8), c_size_t]
libconnect.chirp_connect_as_string.restype = c_char_p
libconnect.chirp_connect_send.argtypes = [c_void_p, POINTER(c_uint8), c_size_t]
libconnect.chirp_connect_send.restype = c_uint8


def get_unique_id():
    """ Get a unique identifier for this device """
    config = configparser.ConfigParser()
    path = os.path.join(os.path.expanduser('~'), '.chirp.ini')
    config.read(path)
    if not config.has_section('default'):
        config.add_section('default')
        config.set('default', 'client_id', str(uuid.uuid4()))
        with open(path, 'w') as cf:
            config.write(cf)
    return config['default']['client_id']


class _Payload(bytearray):
    """
    A Chirp Payload.
    Subclass of built in type - bytearray.
    """
    def __init__(self, sdk, *args, **kwargs):
        self._sdk = sdk
        self.max_length = self._sdk.max_payload_length
        super(_Payload, self).__init__(*args, **kwargs)
        if len(self) > self.max_length:
            raise ValueError('Payload has a maximum length of {}'.format(self.max_length))

    def __str__(self):
        return self._sdk.as_string(self)

    def __eq__(self, other):
        return list(self) == list(other)

    def isvalid(self):
        return self._sdk.is_valid(self)

    def extend(self, data):
        self.max_length = self._sdk.max_payload_length
        if len(self) + len(data) > self.max_length:
            raise ValueError('Payload has a maximum length of {}'.format(self.max_length))
        super(_Payload, self).extend(data)


class CallbackSet(object):
    """
    An abstract class to implement callback methods.

    The ChirpConnectSDK will call these methods when it is
    changing state, or is sending/receiving data. To handle
    data transmission you should inherit this, and assign
    it to the `callbacks` attribute of the SDK after loading
    the config.
    """
    def on_state_changed(self, old, new):
        """
        Called when the state of Chirp Library has changed.

        Args:
            old (str): Previous state
            new (str): Current state
        """
        pass

    def on_sending(self, payload, channel):
        """
        Called when the Chirp Library has begun sending data.

        Args:
            payload (Payload): The Chirp Payload (bytearray)
            channel (int): The channel on which the data is being sent
        """
        pass

    def on_sent(self, payload, channel):
        """
        Called when the Chirp Library has sent the whole payload.

        Args:
            payload (Payload): The Chirp Payload (bytearray)
            channel (int): The channel on which the data is sent
        """
        pass

    def on_receiving(self, channel):
        """
        Called when the Chirp Library hears the start of a payload.

        Args:
            channel (int): The channel on which the data is being received
        """
        pass

    def on_received(self, payload, channel):
        """
        Called when the Chirp Library has decoded a full payload.
        Note: A length of 0 indicates a failed decode.

        Args:
            payload (Payload): The Chirp Payload (bytearray)
            channel (int): The channel on which the data has being received
        """
        pass


class AudioSet(object):
    """
    An abstract class for overriding the default audio layer.

    The ChirpConnectSDK uses `sounddevice` to handle audio streams.
    If you need to provide your own audio layer then you should
    subclass this, and assign it to the `audio` attribute of
    the SDK before loading the config.

    The Audio layer must provide methods to start/stop the
    input/output streams, and close the audio engine. These
    will be called by SDK appropriately.

    Further documentation can be found in the Advanced section
    at https://developers.chirp.io/docs/getting-started/python
    """
    def __init__(self, sdk):
        pass

    def start(self, send=True, receive=True):
        """
        Start the audio engine.

        Args:
            send (bool): Enable output audio
            receive (bool): Enable input audio
        """
        pass

    def stop(self):
        """
        Stop the audio engine.
        """
        pass

    def close(self):
        """
        Close the audio engine.
        """
        pass


class ChirpConnect(object):
    """
    The Chirp Connect SDK.

    This is the main wrapper around the Chirp Connect Library.
    It exposes methods to send arbitrary payloads, and receive
    decoded chirps from the audio layer. Received data is handled
    in callback methods using the `CallbackSet`.

    A full getting started guide for the Chirp Connect SDK can
    be found at https://developers.chirp.io/docs/getting-started/python

    You can set the SDK to run in debug mode by setting the flag true.
    This will write any input data from the microphone to a wav file
    called chirp_audio.wav.

    Args:
        key (str): Chirp application key
        secret (str): Chirp application secret
        config (str): Path to config file or config string
        debug (bool): Enter debug mode
    """
    def __init__(self, key, secret, config=None, debug=False):
        """
        Instantiate the SDK with an application key, secret, and config.
        These can be found at admin.chirp.io.
        """
        self.audio = None
        self.debug = debug
        self.key = key
        self.secret = secret
        self._sdk = new_chirp_connect(self.key, self.secret)
        self._uid = get_unique_id()
        self.callbacks = CallbackSet()
        self._callback_set = ChirpConnectCallbackSet(self)
        self._set_seed()
        if config:
            self.load_config(config) if os.path.exists(config) else self.set_config(config)
        else:
            self.set_config_from_network()
        if not self.is_offline:
            create_instantiate(self.key, self.secret, self._uid)

    def __str__(self):
        """
        Convenience method to output Chirp Connect Library version information.

        Returns:
            dict: Version information for Chirp
                connect/core/crypto libraries.
        """
        version = self.version['connect']
        return 'Chirp Connect {pyversion} [{cversion} {cbuild}] initialised.'.format(
            pyversion=__version__, cversion=version['version'], cbuild=version['build']
        )

    @property
    def version(self):
        """
        Get version information from the Chirp Connect Library.

        Returns:
            dict: Version information for Chirp
                connect/core/crypto libraries.
        """
        v = libconnect.chirp_connect_get_chirp_versions()
        return {
            'connect': {
                'name': v.connect.name.decode(),
                'version': v.connect.version.decode(),
                'build': v.connect.build.decode()
            },
            'core': {
                'name': v.core.name.decode(),
                'version': v.core.version.decode(),
                'build': v.core.build.decode()
            },
            'crypto': {
                'name': v.crypto.name.decode(),
                'version': v.crypto.version.decode(),
                'build': v.crypto.build.decode()
            }
        }

    def close(self):
        """
        Closes sounddevice streams and frees memory in the
        Chirp Connect Library.

        Raises:
            ConnectError: If an error occurs when freeing memory.
        """
        if self.audio:
            self.audio.close()
        if self._sdk:
            del_chirp_connect()

    def _call(self, fn, *args):
        """
        Internal method to make a call to C SDK functions that return
        an error code. If an error occurs, then get the associated
        error message and raise a ConnectError with it.

        Raises:
            ConnectError: If an error occurs.
        """
        rc = fn(*args)
        if rc != 0:
            err = libconnect.chirp_connect_error_code_to_string(rc)
            raise ConnectError(err.decode(), code=rc)

    # -- Getters & Setters

    def load_config(self, filename):
        """
        Load a config from file, and initialise the Chirp Connect Library.

        This must be done after immediately after instantiating the SDK.
        Unless you are implementing your own audio layer, then the audio
        attribute must be set first.

        Args:
            filename (str): The path to the config file

        Raises:
            ConnectError: If the config is invalid or there is an error
                setting the callbacks.
        """
        with open(filename, 'r') as config:
            self.set_config(config.read())

    def set_config(self, config):
        """
        Initialise the Chirp Connect Library with a config.

        This sets up the config and the internal callback set.
        If using the default sounddevice audio layer, then this is instantiated here.

        Args:
            config (str): config string

        Raises:
            ConnectError: If the config is invalid or there is an error setting
                the callbacks.
        """
        if not self.audio:
            self.audio = Audio(self)
        try:
            chirp_connect_set_config(config)
        except UnicodeDecodeError:
            raise ValueError('This doesn\'t look like a valid config')
        except _ConnectError as e:
            raise type(e)(str(e) + '\nEnsure you are passing in a valid config string or file path')

        self._call(libconnect.chirp_connect_set_callbacks, self._sdk, self._callback_set)
        self._max_payload_length = self.max_payload_length

    def set_config_from_network(self, name=None):
        """
        Retrieve a config for your application, and initialise the
        Chirp Connect Library. If no config name is specified, then
        the default config is used.

        Args:
            name (str): config name, default is used if not set

        Raises:
            ConnectNetworkError: If the request to retrieve the config fails.
            ConnectError: If the config is invalid or there is an error setting
            the callbacks.
        """
        config = get_config_from_network(self.key, self.secret, name)
        if config:
            self.set_config(config)

    def _set_seed(self, seed=None):
        """ Set the seed for the random generator """
        seed = randrange(2 ** 32 - 1) if seed is None else seed
        libconnect.chirp_connect_set_random_seed(self._sdk, seed)

    def set_callbacks(self, callbacks):
        """
        Set the callback set, used for consistency between platforms.

        Args:
            callbacks: (CallbackSet): Callback set

        Raises:
            ConnectError: If there is an error setting the callbacks.
        """
        self._call(libconnect.chirp_connect_set_callbacks, self._sdk, self._callback_set)
        self.callbacks = callbacks

    @property
    def is_offline(self):
        """
        bool: Returns true if configured in offline mode.
        """
        return libconnect.chirp_connect_is_offline_mode(self._sdk)

    @property
    def volume(self):
        """
        float: Property to get/set the SDK volume.
        """
        return round(libconnect.chirp_connect_get_volume(self._sdk), 2)

    @volume.setter
    def volume(self, level):
        libconnect.chirp_connect_set_volume(self._sdk, float(level))

    @property
    def sample_rate(self):
        """
        int: Property to get/set the SDK sample rate.
        """
        rv = libconnect.chirp_connect_get_sample_rate(self._sdk)
        if rv == 0:
            raise ConnectError("Couldn't get sample rate (Connect not initialised?)")
        return rv

    @sample_rate.setter
    def sample_rate(self, rate):
        self._call(libconnect.chirp_connect_set_sample_rate, self._sdk, rate)

    @property
    def state(self):
        """
        int: Property to get the current state of the SDK.

        Returns either
            chirpsdk.CHIRP_CONNECT_STATE_NOT_CREATED
            chirpsdk.CHIRP_CONNECT_STATE_STOPPED
            chirpsdk.CHIRP_CONNECT_STATE_PAUSED
            chirpsdk.CHIRP_CONNECT_STATE_RUNNING
            chirpsdk.CHIRP_CONNECT_STATE_SENDING
            chirpsdk.CHIRP_CONNECT_STATE_RECEIVING
        """
        return libconnect.chirp_connect_get_state(self._sdk)

    @property
    def auto_mute(self):
        """
        bool: Property to get/set the auto mute function of the SDK.

        Auto-mute prevents the SDK from hearing its own chirps when
        it is sending data. This should be True in almost all cases.
        """
        return libconnect.chirp_connect_get_auto_mute(self._sdk)

    @auto_mute.setter
    def auto_mute(self, mute):
        libconnect.chirp_connect_set_auto_mute(self._sdk, bool(mute))

    @property
    def protocol_name(self):
        """
        str: Property to get the configured protocol name.
        """
        return libconnect.chirp_connect_get_protocol_name(self._sdk)

    @property
    def protocol_version(self):
        """
        int: Returns the current protocol version.
        """
        return libconnect.chirp_connect_get_protocol_version(self._sdk)

    @property
    def expiry(self):
        """
        datetime: Returns the expiry date of the current configuration.
        """
        dt = libconnect.chirp_connect_get_expiry_time(self._sdk)
        return datetime.utcfromtimestamp(dt)

    def get_duration(self, length):
        """
        float: Returns the duration of a chirp of `length` bytes in seconds.
        """
        return round(libconnect.chirp_connect_get_duration_for_payload_length(self._sdk, length), 3)

    @property
    def channel_count(self):
        """
        int: Returns the number of channels supported by the current configuration.
        """
        return libconnect.chirp_connect_get_channel_count(self._sdk)

    @property
    def transmission_channel(self):
        """
        int: Property to get/set the transmission channel on which data is sent.
        """
        return libconnect.chirp_connect_get_transmission_channel(self._sdk)

    @transmission_channel.setter
    def transmission_channel(self, channel):
        libconnect.chirp_connect_set_transmission_channel(self._sdk, int(channel))

    def get_state_for_channel(self, channel):
        """
        int: Property to get the current state of a channel.

        Returns either
            chirpsdk.CHIRP_CONNECT_STATE_NOT_CREATED
            chirpsdk.CHIRP_CONNECT_STATE_STOPPED
            chirpsdk.CHIRP_CONNECT_STATE_PAUSED
            chirpsdk.CHIRP_CONNECT_STATE_RUNNING
            chirpsdk.CHIRP_CONNECT_STATE_SENDING
            chirpsdk.CHIRP_CONNECT_STATE_RECEIVING
        """
        return libconnect.chirp_connect_get_state_for_channel(self._sdk, int(channel))

    # -- States

    def start(self, send=True, receive=True):
        """
        Start the SDK running.

        The SDK can be configured to run in send mode, receive mode
        or both. The default is send and receive mode.
        This also calls the `start` method in the audio layer.

        Args:
            send (bool): Enable send mode
            receive (bool): Enable receive mode

        Raises:
            ConnectError: If there is an error starting the
                Chirp Connect Library.
        """
        self._call(libconnect.chirp_connect_start, self._sdk)
        if self.audio:
            self.audio.start(send, receive)

    def pause(self, state):
        """
        Pause the SDK and audio streams.

        Args:
            state (bool): True to pause, False to resume

        Raises:
            ConnectError: If there is an error pausing the
                Chirp Connect Library.
        """
        if self.audio:
            self.audio.stop() if state else self.audio.start()
        self._call(libconnect.chirp_connect_pause, self._sdk, bool(state))

    def stop(self):
        """
        Stop the SDK and audio streams.

        Raises:
            ConnectError: If there is an error stopping the
                Chirp Connect Library.
        """
        if self.audio:
            self.audio.stop()
        self._call(libconnect.chirp_connect_stop, self._sdk)

    # -- Callbacks

    def trigger_callbacks(self, payload):
        """
        Trigger the callback methods with a payload.

        Args:
            payload (bytearray): The data for the simulation.
        """
        data = (c_uint8 * len(payload))(*payload)
        libconnect.chirp_connect_trigger_callbacks(self._sdk, data, len(data))

    # -- Processing

    def process_input(self, input):
        """
        Process input data only. Used for receive only mode.

        An input array of floating-point audio samples should be passed in.
        The Chirp Connect Library will trigger the callback methods
        if any chirps are decoded. The input data must contain values
        ranging from -1 to +1, and must be sampled at the same rate as
        the SDK's sample_rate property.

        Args:
            input (:obj:`list` of :obj:`float`): The input audio data.

        Raises:
            ConnectError: If there is an error processing the data.
        """
        chirp_connect_process_input(input)

    def process_output(self, output):
        """
        Process output data only. Used for send only mode.

        An output array should be passed by reference.
        The Chirp Connect Library will populate the
        output buffer with any data to be sent.

        Args:
            output (:obj:`list` of :obj:`float`): The output buffer to be populated.

        Raises:
            ConnectError: If there is an error processing the data.
        """
        chirp_connect_process_output(output)

    def process_shorts_input(self, input):
        """
        Process input data only. Used for receive only mode.

        An input array of 16bit signed integer audio samples should be passed in.
        The Chirp Connect Library will trigger the callback methods
        if any chirps are decoded. The input data must contain values
        ranging from -32767 to +32767, and must be sampled at the same rate as
        the SDK's sample_rate property.

        Args:
            input (:obj:`list` of :obj:`int`): The input audio data.

        Raises:
            ConnectError: If there is an error processing the data.
        """
        chirp_connect_process_shorts_input(input)

    def process_shorts_output(self, output):
        """
        Process output data only. Used for send only mode.

        An output array should be passed by reference.
        The Chirp Connect Library will populate the
        output buffer with any data to be sent.

        Args:
            output (:obj:`list` of :obj:`int`): The output buffer to be populated.

        Raises:
            ConnectError: If there is an error processing the data.
        """
        chirp_connect_process_shorts_output(output)

    # -- Payload

    @property
    def max_payload_length(self):
        """
        int: Property to get the max payload length available with the current configuration.
        """
        return int(libconnect.chirp_connect_get_max_payload_length(self._sdk))

    def new_payload(self, data=None):
        """
        Generate an arbitrary Chirp Payload.
        A Chirp Payload is subclass of the built in type (bytearray).
        It will accept the same data types as a bytearray, ie. an integer,
        a string or an array of bytes.

        Args:
            data: The payload data

        Returns:
            payload (Payload): The Chirp Payload (bytearray)

        Raises:
            ValueError: If the data length is longer than the max payload length.
                This is set by the config. Or if there is no data.
        """
        data = [] if data is None else data
        if len(data) == 0:
            raise ValueError('Cannot generate a null payload')
        return _Payload(self, data)

    def random_payload(self, length=None):
        """
        Generate a random Chirp Payload of `length` bytes
        If no length is specified, then a payload of a random
        length (up to the max_payload_length) is generated.

        Args:
            length: The required length of the payload (optional)

        Returns:
            Payload: A Chirp Payload containing random data (bytearray)
        """
        length = 0 if length is None else length
        lenptr = pointer(c_size_t(length))
        payload = libconnect.chirp_connect_random_payload(self._sdk, lenptr)
        return _Payload(self, payload[:lenptr.contents.value])

    def is_valid(self, payload):
        """
        Check if a payload is valid.

        Args:
            payload (Payload): The Chirp Payload (bytearray)

        Returns:
            bool: True if valid.
        """
        data = (c_uint8 * len(payload))(*payload)
        rc = libconnect.chirp_connect_is_valid(self._sdk, data, len(data))
        return rc == 0

    def as_string(self, payload):
        """
        Returns the string representation of the Chirp Payload.

        Args:
            payload (Payload): The Chirp Payload (bytearray)

        Returns:
            str: The string representation of the Chirp Payload.
        """
        data = (c_uint8 * len(payload))(*payload)
        return str(libconnect.chirp_connect_as_string(self._sdk, data, len(payload)).decode())

    def send(self, payload):
        """
        Send a Chirp Payload.
        This will pass the payload data to the Chirp Connect Library, which will then be
        processed into the output buffer in the `process` functions.

        Args:
            payload (Payload): The Chirp Payload (bytearray)

        Raises:
            ConnectError: If there is an error sending the data.
        """
        data = (c_uint8 * len(payload))(*payload)
        self._call(libconnect.chirp_connect_send, self._sdk, data, len(data))
        if not self.is_offline:
            create_send(
                self.key, self.secret, self._uid, len(data),
                self.protocol_name, self.protocol_version)



# ------------------------------------------------------------------------
#
#  This file is part of the Chirp Connect Python SDK.
#  For full information on usage and licensing, see http://chirp.io/
#
#  Copyright (c) 2011-2018, Asio Ltd.
#  All rights reserved.
#
# ------------------------------------------------------------------------

__title__ = 'chirp'
__version__ = '3.3.1'
__author__ = 'Asio Ltd.'
__license__ = 'License :: Other/Proprietary License'
__copyright__ = 'Copyright 2011-2018, Asio Ltd.'


from ctypes import CDLL, RTLD_GLOBAL
import ctypes.util
import os
import platform

# Explicitly load libgcc as ctypes on RPi doesn't dynamically link it.
gcc = CDLL(ctypes.util.find_library('gcc_s'), mode=RTLD_GLOBAL)

# ------------------------------------------------------------------------
# Locate the libchirp-connect-shared library in order of preference:
#  - the locally-stored library from the libraries subdirectory
#  - the system-wide installation (macOS only)
# To install, download the appropriate Chirp Connect library for your
# operating system and place in the chirp/libraries subdirectory.
# ------------------------------------------------------------------------
try:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    if platform.system() == 'Darwin':
        library_path = os.path.join(dir_path, 'libraries', 'darwin', 'libchirp-connect-shared.dylib')
    elif platform.system() == 'Linux':
        if os.uname()[4].startswith('arm'):
            library_path = os.path.join(dir_path, 'libraries', 'rpi', 'libchirp-connect-shared.so')
        else:
            library_path = os.path.join(dir_path, 'libraries', 'linux', 'libchirp-connect-shared.so')
    else:
        raise RuntimeError('The Chirp Python SDK is not configured for %s platforms. ' \
                           'Please contact contact@chirp.io for custom builds' % platform)
    libconnect = CDLL(library_path)
except OSError:
    library_path = ctypes.util.find_library('chirp-connect-shared')
    if not library_path:
        raise Exception("Couldn't find libchirp-connect-shared. Please install the Chirp C SDK.")
    libconnect = CDLL(library_path)

from .connect import *
from _connect import (
    CHIRP_CONNECT_STATE_STOPPED,
    CHIRP_CONNECT_STATE_PAUSED,
    CHIRP_CONNECT_STATE_RUNNING,
    CHIRP_CONNECT_STATE_SENDING,
    CHIRP_CONNECT_STATE_RECEIVING,
    CHIRP_CONNECT_STATE_NOT_CREATED,
    CHIRP_SDK_BUFFER_SIZE
)
from .exceptions import *



# ------------------------------------------------------------------------
#
#  This file is part of the Chirp Connect Python SDK.
#  For full information on usage and licensing, see https://chirp.io/
#
#  Copyright (c) 2011-2018, Asio Ltd.
#  All rights reserved.
#
# ------------------------------------------------------------------------

import argparse
import sys
import time

from chirpsdk import ChirpConnect, CallbackSet, CHIRP_CONNECT_STATE


class Callbacks(CallbackSet):

    def on_state_changed(self, previous_state, current_state):
        """ Called when the SDK's state has changed """
        print("State changed from {} to {}".format(
            CHIRP_CONNECT_STATE.get(previous_state),
            CHIRP_CONNECT_STATE.get(current_state)))

    def on_sending(self, payload, channel):
        """ Called when a chirp has started to be transmitted """
        print('Sending: ' + str(list(payload)))

    def on_sent(self, payload, channel):
        """ Called when the entire chirp has been sent """
        print('Sent data')
        print(list(payload))

    def on_receiving(self, channel):
        """ Called when a chirp frontdoor is detected """
        print('Receiving data')

    def on_received(self, payload, channel):
        """
        Called when an entire chirp has been received.
        Note: A length of 0 indicates a failed decode.
        """
        if payload is None:
            print('Decode failed!')
        else:
            print('Received:' + str(list(payload)))


def main(app_key, app_secret, app_config,
         input_device, output_device,
         block_size, sample_rate):

    # Initialise ConnectSDK
    sdk = ChirpConnect(app_key, app_secret, app_config)
    print(str(sdk))
    print(sdk.audio.query_devices())

    # Configure audio
    sdk.audio.input_device = input_device
    sdk.audio.output_device = output_device
    sdk.audio.block_size = block_size
    sdk.sample_rate = sample_rate

    # Set callback functions
    sdk.set_callbacks(Callbacks())

    # Generate random payload and send
    payload = sdk.random_payload()
    sdk.start(send=True, receive=True)
    sdk.send(payload)

    try:
        # Process audio streams
        while True:
            time.sleep(0.1)
            sys.stdout.write('.')
            sys.stdout.flush()
    except KeyboardInterrupt:
        print('Exiting')

    sdk.stop()
    sdk.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Chirp Connect SDK Example',
        epilog='Sends a random chirp payload, then continuously listens for chirps'
    )
    parser.add_argument('key', help='Chirp application key')
    parser.add_argument('secret', help='Chirp application secret')
    parser.add_argument('-c', help='Path to config file (optional)')
    parser.add_argument('-i', type=int, default=None, help='Input device index (optional)')
    parser.add_argument('-o', type=int, default=None, help='Output device index (optional)')
    parser.add_argument('-b', type=int, default=0, help='Block size (optional)')
    parser.add_argument('-s', type=int, default=44100, help='Sample rate (optional)')
    args = parser.parse_args()

    main(args.key, args.secret, args.c, args.i, args.o, args.b, args.s)



#!/usr/bin/env python

# ------------------------------------------------------------------------
#
#  This file is part of the Chirp Connect Python SDK.
#  For full information on usage and licensing, see http://chirp.io/
#
#  Copyright (c) 2011-2018, Asio Ltd.
#  All rights reserved.
#
# ------------------------------------------------------------------------

import os
import re
import platform
from setuptools import setup, Extension

vstr = open('chirpsdk/__init__.py', 'r').read()
regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
version = re.search(regex, vstr, re.M)

if platform.system() == 'Darwin':
    arch = 'darwin'
elif platform.system() == 'Linux':
    arch = 'rpi' if os.uname()[4].startswith('arm') else 'linux'
else:
    raise RuntimeError('The Chirp Python SDK is not configured for %s platforms. ' \
                       'Please contact contact@chirp.io for custom builds' % platform)

connect = Extension('_connect',
                    sources=['chirpsdk/_connect.c'],
                    include_dirs=['./chirpsdk', './chirpsdk/include'],
                    library_dirs=['./chirpsdk/libraries/' + arch],
                    runtime_library_dirs=['$ORIGIN/chirpsdk/libraries/' + arch],
                    libraries=['chirp-connect-shared'])

setup(
    name='chirpsdk',
    version=version.group(1),
    description='Chirp Connect Python SDK',
    long_description='The Chirp Connect Python SDK enables the user to send '
                     'and receive data using the device\'s microphone and speaker.',
    license='License :: Other/Proprietary License',
    author='Asio Ltd.',
    author_email='developers@chirp.io',
    url='https://developers.chirp.io',
    packages=['chirpsdk', 'tests', 'bin'],
    ext_modules=[connect],
    install_requires=[
        'sounddevice>=0.3.10', 'pysoundfile>=0.9.0',
        'requests>=2.18.1', 'requests-futures==0.9.7',
        'configparser==3.5.0'
    ],
    include_package_data=True,
    tests_require=['mock==2.0.0'],
    keywords=['sound', 'networking', 'chirp'],
    scripts=[
        'bin/chirp-audio-read', 'bin/chirp-audio-write',
        'bin/chirp-receive', 'bin/chirp-send'
    ],
    test_suite='tests',
    classifiers=[
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Communications',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers'
    ]
)
